Snake
=======

Game Description
----------------
In the Snake game, a snake wanders around on the screen, steered by the a single player using the arrow keys: Up, Down, Left, Right. The player must avoid running off the screen or having the snake run into its own body--if that occurs the game is over. To make the game more challenging, the snake's body gets longer as the game goes on, and it moves faster over time. Use a score counter to keep track of how long the snake has been alive.

Preliminary thinking
--------------------
The snake, most simply, consists of squares on a grid, and we'll keep track of those squares using a small list of cell coordinates. If the snake currently occupies positions (0,1), (0,2), and (0,3), we can keep those locations in list:

    cells = [ [0,1], [0,2], [0,3] ]
    
It probably makes sense to have a Snake class that we can use to manage those cells as the snake moves. That class will have a method `move(direction)` that moves the snake in the indicated direction, where direction 0 = right, 1 = down, 2 = left, and 3 = up.

The cells have some dimension to them in pixels. If we have a screen that's 800 x 600 pixels, each square would have a dimension of (for example) 50 pixels. So a snake cell that's located at [2, 3] would actually be placed at position [2 x 50, 3 x 50] = [100, 150] on the screen. We'll need to figure out a system for managing the spaces on the screen, where each space is made up of a number of pixels.

Description of Program
----------------------

Once we've got the Snake class set up, the main program will create a screen, create a Snake object, and repeatedly draw that snake on the screen, moving it as the game progresses according to the directions given by the player.

We'll have to figure out how to draw the screen, and how to get keyboard input from the player. We'll have to figure out how to make the Snake grow, too--that will probably require a new method in the `Snake` class, and we'll have to figure out how to change the timing of the snake's movement so we can increase the game speed as it goes.

General Outline of Program
--------------------------

    import Snake
    
    def setup():
        # list global variables so draw can use them
        # set screen size
        # set background color
        sc = 50     # the scale of the screen divisions, in pixels
        startX = width / 2 / sc
        startY = height / 2 / sc
        # set initial cell list for snake
        s = Snake.Snake(cells)
        # set initial direction of travel (0-3)
        timeDelay = 1000
        startTime = millis()
        moves = 0
        
    def draw_screen():
        # draw the screen with gridlines
        # draw the snake squares
        
    def draw():
        # list global variables so draw() can use them
        # call draw_screen()
        # Check to see if a key was pressed, and if so,
        # which one. Use that to set the direction
        
        if millis() - startTime > timeDelay:
            s.move(direction)
            # increment move counter
            timeDelay = timeDelay * 0.985  # speed things up!
            startTime = millis()           # restart counter
        

Keypresses
----------
We need to have an "event listener" that checks to see if a key is pressed. The Processing platform has this built in. Here is the syntax of that listener in the form of an `if-else` statement.

    if keyPressed and key == CODED:
        if (keyCode == LEFT):
            direction = 2
        elif (keyCode == RIGHT):
            direction = 0
        elif (keyCode == UP):
            direction = 3
        else: # DOWN
            direction = 1
   

Timing
------
We're going to have the snake move once per second (for example) at the beginning of the game, and have it slowly advance its speed. Although Processing has a `delay` function that we can use to momentarily halt the program, we don't want to use that. We don't want the program to stop accepting input, we just want to keep the snake from moving for a period of time. How to do that?

There is a Processing function called `millis()` that keeps track of the number of milliseconds it has been since the program started running. If we have variables to keep track of a `startTime` and a `delayTime`, we can do the following (pseudocode):

    # At beginning of program:
    # Set startTime to millis()  
    # Set delayTime to 1000
    
    # In the draw() loop:
    # if the current millis() > startTime + delayTime:
        # startTime = millis()    # reset the timer with the current time
        # delayTime = delayTime * 0.97   # reduce the delayTime a small amount
                                         # so speed of moves increases

Losing the Game
---------------
The player loses the game if the snake goes outside the boundaries of the screen or if the snake crosses over onto its own body. We can write a function to check for those situations, and end the game if one of those happens.

    def check_losing(s, sc):
    """Checks to see if the snake has gone off screen
    or crossed over itself, in which case the game is
    over.
    """
    if s.getSnake()[-1] in s.getSnake()[0:-1] or \
       s.getSnake()[-1][0] * sc < 0 or s.getSnake()[-1][0] * sc == width or \
       s.getSnake()[-1][1] * sc < 0 or s.getSnake()[-1][1] * sc == height:
        background(255,0,0)
        noLoop()

Displaying the Score
--------------------
The score can be displayed in the console by using a `print()` statement, but even better is displaying the score onscreen. We can do that by specifying the font, size, and location of the text that we want to display.

In the `setup()` function include the lines

    textSize(24)
    textAlign(CENTER)
    
Find where in the `draw` loop that you want to display the score, and try this:

    text("Steps: " + str(moves), width / 2, height - 20)

Making the Snake Grow
---------------------
One of the convenient ways of making the snake grow (extend its length by one cell) is to do so in the `.move()` function. In addition to sending in a `direction`, parameter, we'll also send in a boolean variable `grow` that will extend the snake's length by one if its `True`.

How can we actually do that? We've been keeping track of the snake's cells in a list. Normally we copy the second cell into the first, the third cell into the second, and then we put a new pair of coordinates in the third cell, depending on which direction the snake is heading.

def move(self, direction):
    """0 = right, 1 = down, 2 = left, 3 = up
    """
    if direction == 0:   # right
        newRow = self.cells[-2][0] + 1
        newCol = self.cells[-2][1]
        self.cells[-1] = [newRow, newCol]
    if direction == 2:   # left
        newRow = self.cells[-2][0] - 1
        newCol = self.cells[-2][1]
        self.cells[-1] = [newRow, newCol]
    if direction == 1:   # down
        newRow = self.cells[-2][0]
        newCol = self.cells[-2][1] + 1
        self.cells[-1] = [newRow, newCol]
    if direction == 3:   # down
        newRow = self.cells[-2][0]
        newCol = self.cells[-2][1] - 1
        self.cells[-1] = [newRow, newCol]
        
In the new version of this, if `grow` is True, we *won't* shift snake cells down. We'll leave them where they are, but append a new cell to the end of the list. (And if `grow` is `False` we'll do just what we did before.)

    if grow:
        self.cells.append([0,0]) # blank cell
    else:
        for i in range(1, len(self.cells)):
            self.cells[i - 1] = self.cells[i]
            
And that's it. We've assigned a blank cell at the end, but that will be replaced with a real location based on the current direction, just as before.
